module ILCProblems

export ILCProblem
export solve!

using ..Experiments
using ..QuadraticProblems

using NamedTrajectories
using TrajectoryIndexingUtils
using QuantumCollocation: iso_to_ket
using QuantumCollocation.Integrators

using LinearAlgebra
using SparseArrays
using Statistics

mutable struct ILCProblem
    Ẑgoal::NamedTrajectory
    Ẑ::NamedTrajectory
    QP::QuadraticProblem
    Ygoal::MeasurementData
    Ȳs::Vector{MeasurementData}
    As::Vector{Matrix{Float64}}
    experiment::AbstractExperiment
    settings::Dict{Symbol, Any}
    bt_dict::Dict{Int, Vector}
    loss::Vector{Float64}

    function ILCProblem(
        Ẑgoal::NamedTrajectory,
        dynamics::Function,
        experiment::QuantumExperiment;
        Qy=1.0,
        Qyf=100.0,
        R=(a=1.0, dda=1.0),
        correction_term=true,
        dynamic_measurements=true,
        verbose=true,
        max_iter=100,
        max_backtrack_iter=10,
        tol=1e-6,
        α=0.5,
        β=0.5,
        norm_p=Inf,
        static_QP=false,
        QP_max_iter=100_000,
        QP_verbose=false,
        QP_linear_solver="mkl pardiso",
        QP_settings=Dict(),
        QP_tol=1e-9,
        mle=true,
        Σ=nothing,
        y_final=nothing,
    )

        @assert R isa NamedTuple "R must be a NamedTuple specifying components to regularize -- e.g. R=(a=1.0, dda=1.0)"

        ILC_settings::Dict{Symbol, Any} = Dict(
            :max_iter => max_iter,
            :max_backtrack_iter => max_backtrack_iter,
            :α => α,
            :β => β,
            :tol => tol,
            :norm_p => norm_p,
            :verbose => verbose,
        )

        QP_kw_settings::Dict{Symbol, Any} = Dict(
            :max_iter => QP_max_iter,
            :verbose => QP_verbose,
            :eps_abs => QP_tol,
            :eps_rel => QP_tol,
            :eps_prim_inf => QP_tol,
            :eps_dual_inf => QP_tol,
            :linsys_solver => QP_linear_solver,
        )

        QP_settings::Dict{Symbol, Any} =
            merge(QP_kw_settings, QP_settings)

        @assert length(dynamics(Ẑgoal[1].data, Ẑgoal[2].data)) ==
            Ẑgoal.dims.states

        f = zₜzₜ₊₁::AbstractVector{<:Real} -> begin
            zₜ = zₜzₜ₊₁[slice(1, Ẑgoal.dim)]
            zₜ₊₁ = zₜzₜ₊₁[slice(2, Ẑgoal.dim)]
            return dynamics(zₜ, zₜ₊₁)
        end

        if static_QP
            println("Setting up static QP...")
            QP = StaticQuadraticProblem(
                Ẑgoal,
                f,
                experiment.g,
                experiment.τs,
                Qy, Qyf, R,
                correction_term,
                dynamic_measurements,
                QP_settings;
                mle=mle
            )
        else
            println("Setting up dynamic QP...")
            QP = DynamicQuadraticProblem(
                Ẑgoal,
                f,
                experiment.g,
                Qy, Qyf, R,
                correction_term,
                QP_settings;
                mle=mle
            )
        end

        Ygoal = measure(
            Ẑgoal.ψ̃,
            experiment.g,
            experiment.τs,
            experiment.ydim
        )

        if !isnothing(y_final)
            Ygoal.ys[end] = y_final
        end

        return new(
            Ẑgoal,
            Ẑgoal,
            QP,
            Ygoal,
            MeasurementData[],
            Matrix{Float64}[],
            experiment,
            ILC_settings,
            Dict{Int, Vector}(),
            Float64[],
        )
    end


    function ILCProblem(
        Ẑgoal::NamedTrajectory,
        dynamics::Function,
        experiment::HardwareExperiment;
        Qy=1.0,
        Qyf=100.0,
        R=(a=1.0, dda=1.0),
        correction_term=true,
        dynamic_measurements=true,
        verbose=true,
        max_iter=100,
        max_backtrack_iter=10,
        tol=1e-6,
        α=0.5,
        β=0.5,
        norm_p=Inf,
        static_QP=false,
        QP_max_iter=100_000,
        QP_verbose=false,
        QP_linear_solver="mkl pardiso",
        QP_settings=Dict(),
        QP_tol=1e-9,
        mle=true,
        Σ=nothing,
        y_final=nothing,
    )

        @assert R isa NamedTuple "R must be a NamedTuple specifying components to regularize -- e.g. R=(a=1.0, dda=1.0)"

        ILC_settings::Dict{Symbol, Any} = Dict(
            :max_iter => max_iter,
            :max_backtrack_iter => max_backtrack_iter,
            :α => α,
            :β => β,
            :tol => tol,
            :norm_p => norm_p,
            :verbose => verbose,
        )

        QP_kw_settings::Dict{Symbol, Any} = Dict(
            :max_iter => QP_max_iter,
            :verbose => QP_verbose,
            :eps_abs => QP_tol,
            :eps_rel => QP_tol,
            :eps_prim_inf => QP_tol,
            :eps_dual_inf => QP_tol,
            :linsys_solver => QP_linear_solver,
        )

        QP_settings::Dict{Symbol, Any} =
            merge(QP_kw_settings, QP_settings)

        @assert length(dynamics(Ẑgoal[1].data, Ẑgoal[2].data, Ẑgoal)) ==
            Ẑgoal.dims.states

        f = zₜzₜ₊₁::AbstractVector{<:Real} -> begin
            zₜ = zₜzₜ₊₁[slice(1, Ẑgoal.dim)]
            zₜ₊₁ = zₜzₜ₊₁[slice(2, Ẑgoal.dim)]
            return dynamics(zₜ, zₜ₊₁)
        end

        if static_QP
            QP = StaticQuadraticProblem(
                Ẑgoal,
                f,
                experiment.g_analytic,
                experiment.τs,
                Qy, Qyf, R,
                correction_term,
                dynamic_measurements,
                QP_settings;
                mle=mle
            )
        else
            QP = DynamicQuadraticProblem(
                Ẑgoal,
                f,
                experiment.g_analytic,
                Qy, Qyf, R,
                correction_term,
                QP_settings;
                mle=mle
            )
        end

        Ygoal = measure(
            Ẑgoal.ψ̃,
            experiment.g_analytic,
            experiment.τs,
            experiment.ydim
        )

        if !isnothing(y_final)
            @assert length(y_final) == experiment.ydim
            Ygoal.ys[end] = y_final
        end

        return new(
            Ẑgoal,
            Ẑgoal,
            QP,
            Ygoal,
            MeasurementData[],
            Matrix{Float64}[],
            experiment,
            ILC_settings,
            Dict{Int, Vector}(),
            Float64[]
        )
    end
end

function fidelity(ψ̃, ψ̃_goal)
    ψ = iso_to_ket(ψ̃)
    ψ_goal = iso_to_ket(ψ̃_goal)
    return abs2(ψ' * ψ_goal)
end

function fidelity(ψ̃)
    ψ = iso_to_ket(ψ̃)
    ketdim = length(ψ)
    ψ_goal = [0.0; 1.0; zeros(ketdim - 2)]
    return abs2(ψ' * ψ_goal)
end

function pops(ψ̃)
    ψ = iso_to_ket(ψ̃)
    return abs2.(ψ)
end



function solve!(
    prob::ILCProblem;
    σ=1.0,
    nominal_correction=true,
    nominal_rollout_integrator=Integrators.fourth_order_pade,
    show_metric=false,
    metric=nothing,
)
    # get the initial controls and add to the stored list of controls
    A = prob.Ẑ.a
    push!(prob.As, A)

    # get the initial measurement and add to the stored list of measurements
    Ȳ = prob.experiment(A, times(prob.Ẑ))
    push!(prob.Ȳs, Ȳ)

    if metric == :full_state
        println("|g1⟩ component measured = ", pops(Ȳ.ys[end])[2])
        println("|g1⟩ component goal     = ", pops(prob.Ygoal.ys[end])[2])
    else
        println("|g1⟩ component measured = ", Ȳ.ys[end][2])
        println("|g1⟩ component goal     = ", prob.Ygoal.ys[end][2])
    end

    # compute the initial error
    ΔY = Ȳ - prob.Ygoal

    # ψ̃goal = prob.Ẑgoal.ψ̃[:, end]


    # display(fidelity(Ȳ.ys[end], ψ̃goal))
    # display(fidelity(Ȳ.ys[end]))
    # display(fidelity(ψ̃goal))

    # println("|⟨ψf|ψgoal⟩|² = ", fidelity(Ȳ.ys[end], ψ̃goal))
    # println("|⟨ψf|g1⟩|²    = ", fidelity(Ȳ.ys[end]))
    # println("|⟨ψgoal|g1⟩|² = ", fidelity(ψ̃goal))

    # initialize the tracked list of normed measurement errors
    ΔyT_norms = [norm(ΔY.ys[end], prob.settings[:norm_p])]

    # initialize the list of tracked fidelities
    # fidelities = [fidelity(Ȳ.ys[end], ψ̃goal)]

    # initialize the list of tracked populations
    # g1_pops = [pops(Ȳ.ys[end])]

    # initialize the iteration counter
    k = 1

    while true

        # check if we've reached the maximum number of ILC iterations
        if k > prob.settings[:max_iter]
            @info "max iterations reached" max_iter = prob.settings[:max_iter]
            return
        end

        # printed ILC iteratation information
        if prob.settings[:verbose]
            println()
            printstyled("iter      = ", k; color=:magenta)
            println()
            # printstyled("⟨|ΔY|⟩  = ", mean([norm(y, prob.settings[:norm_p]) for y in ΔY.ys]); color=:magenta)
            # println()
            printstyled("|ΔY(T)|   = ", norm(ΔY.ys[end], prob.settings[:norm_p]); color=:magenta)
            println()
            if show_metric
                # printstyled("|⟨ψ|ψg⟩|² = ", fidelity(Ȳ.ys[end], ψ̃goal); color=:magenta)
                if metric == :full_state
                    printstyled("|g1⟩ pop  = ", pops(Ȳ.ys[end])[2]; color=:magenta)
                else
                    printstyled("|g1⟩ pop  = ", Ȳ.ys[end][2]; color=:magenta)
                end
            end
            println()
            println()
        end

        # get step direction from QP
        ΔZ = prob.QP(ΔY, prob.Ẑ)

        # mutliply by pre-scaling factor β
        ΔZ *= prob.settings[:β]

        # update nominal trajectory
        prob.Ẑ += ΔZ

        # get the updated controls
        A = prob.Ẑ.a

        # run an experiment with the updated controls
        Ȳ = prob.experiment(A, times(prob.Ẑ))

        # compute the difference between the updated measurements and the goal
        ΔY_next = Ȳ - prob.Ygoal # type: MeasurementData

        # get the vector valued final measurement difference
        ΔyT_next = ΔY_next.ys[end]
        # fidelity_next = fidelity(Ȳ.ys[end], ψ̃goal)
        # g1_pop_next = pops(Ȳ.ys[end])[2]

        # line search phase of ILC
        # TODO: make this modular
        if norm(ΔyT_next, prob.settings[:norm_p]) >= minimum(ΔyT_norms)
        # if fidelity(Ȳ.ys[end], ψ̃goal) <= minimum(fidelities)

            printstyled("   backtracking"; color=:magenta)
            println()
            println()

            # initialize the backtracking iteration counter
            i = 1

            # initialize the list of final measurements during backtracking
            backtrack_yTs = []

            # initialize the list of measurement errors during backtracking
            iter_ΔyT_norms = []

            # initialze the list of fidelities during backtracking
            # iter_fidelities = []


            # backtracking loop
            while norm(ΔyT_next, prob.settings[:norm_p]) >= minimum(ΔyT_norms)
            # while fidelity_next <= maximum(fidelities)

                # check if we've reached the maximum number of backtracking iterations
                if i > prob.settings[:max_backtrack_iter]
                    ΔY = ΔY_next
                    push!(prob.loss, iter_ΔyT_norms[end])
                    println()
                    printstyled("   max backtracking iterations reached"; color=:magenta)
                    println()
                    println()
                    printstyled("iter      = ", k; color=:magenta)
                    println()
                    # printstyled("⟨|ΔY|⟩  = ", mean([norm(y, prob.settings[:norm_p]) for y in ΔY.ys]); color=:magenta)
                    # println()
                    printstyled("|ΔY(T)|   = ", iter_ΔyT_norms[end]; color=:magenta)
                    println()
                    if show_metric
                        # printstyled("|⟨ψ|ψg⟩|² = ", fidelity(Ȳ.ys[end], ψ̃goal); color=:magenta)
                        if metric == :full_state
                            printstyled("|g1⟩ pop  = ", pops(Ȳ.ys[end])[2]; color=:magenta)
                        else
                            printstyled("|g1⟩ pop  = ", Ȳ.ys[end][2]; color=:magenta)
                        end
                    end
                    println()
                    return
                end

                # reset to previous state
                prob.Ẑ -= ΔZ

                # decrease step size by a factor of α
                ΔZ *= prob.settings[:α]

                # take step
                prob.Ẑ += ΔZ

                # get pulse controls
                A = prob.Ẑ.a

                # get the final measurement of the experiment with the updated controls
                yT = prob.experiment(A, times(prob.Ẑ); backtracking=true).ys[end]

                # store the final measurement
                push!(backtrack_yTs, yT)

                # compute the difference between the updated measurement and the goal
                ΔyT_next = yT - prob.Ygoal.ys[end]
                # fidelity_next = fidelity(yT, ψ̃goal)

                # store the norm of the final measurement difference
                push!(iter_ΔyT_norms, norm(ΔyT_next, prob.settings[:norm_p]))
                # push!(iter_fidelities, fidelity_next)

                # printed backtracking iteration information
                println()
                printstyled("   backtracking iter = ", i; color=:magenta)
                println()
                println()
                printstyled("     min |ΔY(T)|   = ", minimum(ΔyT_norms); color=:cyan)
                println()
                printstyled("       |ΔY(T)|     = ", norm(ΔyT_next, prob.settings[:norm_p]); color=:cyan)
                println()
                if show_metric
                # printstyled("    max |⟨ψ|ψg⟩|²  = ", maximum(fidelities); color=:cyan)
                    # println()
                # printstyled("      |⟨ψ|ψg⟩|²    = ", fidelity(yT, ψ̃goal); color=:cyan)
                #     println()
                    # printstyled("    max |g1⟩ pop = ", maximum(fidelities); color=:cyan)
                    println()
                    if metric == :full_state
                        printstyled("      |g1⟩ pop     = ", pops(yT)[2]; color=:cyan)
                    else
                        printstyled("      |g1⟩ pop     = ", yT[2]; color=:cyan)
                    end
                    println()
                end
                println()
                # increment backtracking iteration counter
                i += 1
            end

            # update the list of normed measurement errors
            # with the minimum of the backtracking iteration errors
            push!(ΔyT_norms, minimum(iter_ΔyT_norms))
            # push!(fidelities, maximum(iter_fidelities))

            # store the backtracking per iteration measurement data
            prob.bt_dict[k] = backtrack_yTs

            # remeasure with new controls to get full measurement
            Ȳ = prob.experiment(A, times(prob.Ẑ))

            # compute new error
            ΔY = Ȳ - prob.Ygoal

            # push remeasured norm(ΔY) to tracked errors
            push!(ΔyT_norms, norm(ΔY.ys[end], prob.settings[:norm_p]))
            push!(prob.loss, norm(ΔY.ys[end], prob.settings[:norm_p]))
            # push!(fidelities, fidelity(Ȳ.ys[end], ψ̃goal))
        else
            # if not backtracking, just update the error
            ΔY = ΔY_next

            # push normed final error to tracked errors
            push!(ΔyT_norms, norm(ΔY.ys[end], prob.settings[:norm_p]))
            # push!(fidelities, fidelity(Ȳ.ys[end], ψ̃goal))
        end

        # store the updated measurement data
        push!(prob.Ȳs, Ȳ)

        # store the updated controls
        push!(prob.As, A)

        if nominal_correction
            # get corrected nominal trajectory
            Ψ̃_corrected = nominal_rollout(A, times(prob.Ẑ), prob.experiment; integrator=nominal_rollout_integrator)

            # update nominal trajectory
            update!(prob.Ẑ, :ψ̃, Ψ̃_corrected)
        end

        # increment ILC iteration counter
        k += 1
    end
end

end
