module QuadraticProblems

export QuadraticProblem

export StaticQuadraticProblem
export DynamicQuadraticProblem

using ..Experiments

using NamedTrajectories
using TrajectoryIndexingUtils
using QuantumCollocation
using LinearAlgebra
using SparseArrays
using ForwardDiff
using Zygote
using Einsum
using OSQP

abstract type QuadraticProblem end

struct StaticQuadraticProblem <: QuadraticProblem
    Hreg::SparseMatrixCSC
    A::SparseMatrixCSC
    ∂gs::Vector{Matrix{Float64}}
    ∂²gs::Vector{Array{Float64}}
    ∂g::Function
    ∂²g::Function
    Qy::Float64
    Qyf::Float64
    R::NamedTuple{names, <:Tuple{Vararg{Float64}}} where names
    correction_term::Bool
    dynamic_measurements::Bool
    mle::Bool
    settings::Dict{Symbol, Any}
end

function StaticQuadraticProblem(
    Ẑgoal::NamedTrajectory,
    f::Function,
    g::Function,
    τs::AbstractVector{Int},
    Qy::Float64,
    Qyf::Float64,
    R::NamedTuple{names, <:Tuple{Vararg{Float64}}} where names,
    correction_term::Bool,
    dynamic_measurements::Bool,
    settings::Dict{Symbol, Any};
    mle::Bool=true
)
    @assert :ψ̃ ∈ keys(Ẑgoal.components) "must specify quantum states component as ψ̃"
    @assert all([k ∈ keys(Ẑgoal.components) for k in keys(R)]) "regularization multiplier keys must match trajectory component keys"

    ydim = length(g(Ẑgoal[1].ψ̃))


    # dynamics constraints

    @assert length(f(vec(Ẑgoal.data[:, 1:2]))) == Ẑgoal.dims.states

    ∂f(zz) = Zygote.jacobian(f, zz)[1]

    ∂F = build_dynamics_constraint_jacobian(Ẑgoal, ∂f)


    # measurement constraints

    ∂g(x) = Zygote.jacobian(g, x)[1]

    function ∂²g(x)
        H = Zygote.jacobian(u -> vec(∂g(u)), x)[1]
        return reshape(H, ydim, Ẑgoal.dims.ψ̃, Ẑgoal.dims.ψ̃)
    end

    ∂gs = Matrix[]

    for τ ∈ τs
        push!(∂gs, ∂g(Ẑgoal[τ].ψ̃))
    end

    ∂²gs = Array[]

    for τ ∈ τs
        push!(∂²gs, ∂²g(Ẑgoal[τ].ψ̃))
    end


    # bounds constraints
    B = build_bounds_constraint_matrix(Ẑgoal)

    # initial/final value constraints
    C = build_initial_and_final_constraint_matrix(Ẑgoal)

    # full constraint matrix
    A = sparse_vcat(∂F, B, C)

    # hessian of regularization objective terms
    Hreg = build_regularization_hessian(R, Ẑgoal)

    return StaticQuadraticProblem(
        Hreg,
        A,
        ∂gs,
        ∂²gs,
        ∂g,
        ∂²g,
        Qy,
        Qyf,
        R,
        correction_term,
        dynamic_measurements,
        mle,
        settings
    )
end

function (QP::StaticQuadraticProblem)(
    ΔY::MeasurementData,
    Ẑ::NamedTrajectory
)

    # constraint values
    lb, ub = build_constraints(Ẑ)

    if QP.mle
        Σinvs = [inv(Σ) for Σ ∈ ΔY.Σs]
        if QP.dynamic_measurements
            Hmle, ∇ = build_mle_hessian_and_gradient(
                ΔY,
                QP.∂g,
                QP.∂²g,
                QP.Qy,
                QP.Qyf,
                Σinvs,
                Ẑ.dims,
                Ẑ.components.ψ̃,
                Ẑ.T,
                QP.correction_term
            )
        else
            Hmle, ∇ = build_mle_hessian_and_gradient(
                ΔY,
                QP.∂gs,
                QP.∂²gs,
                QP.Qy,
                QP.Qyf,
                Σinvs,
                Ẑ.dims,
                Ẑ.components.ψ̃,
                Ẑ.T,
                QP.correction_term
            )
        end

        H = Hmle + QP.Hreg
        A = QP.A
    else
        H = Hreg
        ∂G, ∂G_cons = build_measurement_constraint_jacobian(
            ΔY, QP.∂gs, QP.∂²gs, QP.dims, QP.correction_term
        )
        A = sparse_vcat(∂G, QP.A)
        ub = vcat(∂G_cons, ub)
        lb = vcat(∂G_cons, lb)
    end

    model = OSQP.Model()

    OSQP.setup!(
        model;
        P=H,
        A=A,
        q=QP.mle ? ∇ : nothing,
        l=lb,
        u=ub,
        QP.settings...
    )

    results = OSQP.solve!(model)

    if results.info.status != :Solved
        @warn "OSQP did not solve the problem. status: $(results.info.status)"
        println()
    end

    ΔZ = NamedTrajectory(results.x, Ẑ)

    return ΔZ
end


struct DynamicQuadraticProblem <: QuadraticProblem
    ∂f::Function
    ∂g::Function
    ∂²g::Function
    Qy::Float64
    Qyf::Float64
    R::NamedTuple{names, <:Tuple{Vararg{Float64}}} where names
    C::SparseMatrixCSC{Float64, Int64}
    Hreg::SparseMatrixCSC{Float64, Int64}
    correction_term::Bool
    mle::Bool
    settings::Dict{Symbol, Any}
end

function DynamicQuadraticProblem(
    Z::NamedTrajectory,
    f::Function,
    g::Function,
    Qy::Float64,
    Qf::Float64,
    R::NamedTuple{names, <:Tuple{Vararg{Float64}}} where names,
    correction_term::Bool,
    settings::Dict{Symbol, Any};
    mle::Bool=true
)
    @assert :ψ̃ ∈ keys(Z.components) "must specify quantum states component as ψ̃"

    @assert all([k ∈ keys(Z.components) for k ∈ keys(R)]) "regularization multiplier keys must match trajectory component keys"

    @assert length(f(vec(Z.data[:, 1:2]))) == Z.dims.states

    ∂f(zz) = ForwardDiff.jacobian(f, zz)

    # Symbolics.@variables zz[1:2 * Z.dim]

    # zz = collect(zz)

    # ∂f_symb = Symbolics.sparsejacobian(f(zz), zz; simplify=true)
    # ∂f_expr = Symbolics.build_function(∂f_symb, zz; expression=true)
    # ∂f = zz -> sparse(eval(∂f_expr[1])(zz))

    ∂g(x) = ForwardDiff.jacobian(g, x)

    # Symbolics.@variables ψ̃[1:Z.dims.ψ̃]

    # ψ̃ = collect(ψ̃)

    # ∂g_symb = Symbolics.sparsejacobian(g(ψ̃), ψ̃; simplify=true)
    # ∂g_expr = Symbolics.build_function(∂g_symb, ψ̃; expression=true)
    # ∂g = ψ̃ -> sparse(eval(∂g_expr[1])(ψ̃))

    ydim = length(g(Z[1].ψ̃))

    function ∂²g(x)
        H = ForwardDiff.jacobian(u -> vec(∂g(u)), x)
        return reshape(H, ydim, Z.dims.ψ̃, Z.dims.ψ̃)
    end

    # ∂²g_symb = Symbolics.sparsejacobian(vec(∂g(ψ̃)), ψ̃; simplify=true)
    # ∂²g_expr = Symbolics.build_function(∂²g_symb, ψ̃; expression=true)
    # ∂²g = ψ̃ -> reshape(eval(∂²g_expr[1])(ψ̃), ydim, Z.dims.ψ̃, Z.dims.ψ̃)

    println("Building bounds constraint matrix...")
    C_bounds = build_bounds_constraint_matrix(Z)

    println("Building initial/final value constraint matrix...")
    C_init_final = build_initial_and_final_constraint_matrix(Z)

    println("Building initial wavefunction constraint matrix...")
    C_ψ̃₁ = build_initial_wavefunction_constraint_matrix(Z)

    C = sparse_vcat(C_bounds, C_init_final, C_ψ̃₁)

    println("Building regularization hessian...")
    Hreg = build_regularization_hessian(R, Z)

    return DynamicQuadraticProblem(
        ∂f,
        ∂g,
        ∂²g,
        Qy,
        Qf,
        R,
        C,
        Hreg,
        correction_term,
        mle,
        settings
    )
end

function (QP::DynamicQuadraticProblem)(
    ΔY::MeasurementData,
    Ẑ::NamedTrajectory
)
    model = OSQP.Model()

    # get the dynamics constraint jacobian
    println("Building dynamics constraint jacobian...")
    ∂F = build_dynamics_constraint_jacobian(Ẑ, QP.∂f)

    # full constraint matrix
    A = sparse_vcat(∂F, QP.C)

    # get the constraint bounds
    println("Building constraint bounds...")
    lb, ub = build_constraints(Ẑ)

    if QP.mle

        # get the inverses of the measurment covariances
        Σinvs = [inv(Σ) for Σ ∈ ΔY.Σs]

        # get the hessian and gradient of the mle objective term
        println("Building mle hessian and gradient...")
        Hmle, ∇ = build_mle_hessian_and_gradient(
            ΔY,
            Ẑ,
            QP.∂g,
            QP.∂²g,
            QP.Qy,
            QP.Qyf,
            Σinvs,
            Ẑ.dims,
            Ẑ.components.ψ̃,
            Ẑ.T,
            QP.correction_term
        )

        # build the full hessian
        H = Hmle + QP.Hreg
    else
        ∂G, ∂G_cons = build_measurement_constraint_jacobian(
            Ẑ, ΔY, QP.∂g, QP.∂²g, QP.dims, QP.correction_term
        )
        A = sparse_vcat(∂G, A)
        lb = vcat(∂G_cons, lb)
        ub = vcat(∂G_cons, ub)
        H = QP.Hreg
    end

    println("Setting up OSQP model...")
    OSQP.setup!(
        model;
        P=H,
        A=A,
        q=QP.mle ? ∇ : nothing,
        l=lb,
        u=ub,
        QP.settings...
    )

    println("Solving OSQP model...")
    results = OSQP.solve!(model)

    if results.info.status != :Solved
        @warn "OSQP did not solve the problem, status: $(results.info.status)"
    end

    ΔZ = NamedTrajectory(results.x, Ẑ)

    return ΔZ
end


"""
    build_regularization_hessian(
        R::NamedTuple{names, Tuple{Vararg{Float64}}} where names,
        Z::NamedTrajectory
    )

Builds the hessian matrix for the regularization objective terms.
"""
@inline function build_regularization_hessian(
    R::NamedTuple{names, <:Tuple{Vararg{Float64}}} where names,
    Z::NamedTrajectory
)
    Hₜ = blockdiag([
        k ∈ keys(R) ?
            R[k] * sparse(I(Z.dims[k])) :
            spzeros(Z.dims[k], Z.dims[k])
        for k ∈ Z.names
    ]...)
    H = kron(sparse(I(Z.T)), Hₜ)
    return H
end


"""
    build_mle_hessian_and_gradient()

Builds the hessian matrix and gradient vector for the MLE objective term.
"""
@inline function build_mle_hessian_and_gradient(
    ΔY::MeasurementData,
    ∂gs::Vector{Matrix{Float64}},
    ∂²gs::Vector{Array{Float64}},
    Qy::Float64,
    Qyf::Float64,
    Σinvs::Vector{<:AbstractMatrix},
    dims::NamedTuple,
    measured_components::AbstractVector{Int},
    T::Int, # number of time steps
    correction_term::Bool
)
    M = length(ΔY.τs)
    zdim = dims.states + dims.controls

    Hmle = spzeros(zdim * T, zdim * T)

    ∇ = zeros(zdim * T)

    for i = 1:M

        τᵢ = ΔY.τs[i]

        ∂gᵢ = ∂gs[i]

        if correction_term
            ∂²gᵢ = ∂²gs[i]
            ϵ̂ᵢ = pinv(∂gᵢ) * ΔY.ys[i]
            @einsum ∂gᵢ[j, k] += ∂²gᵢ[j, k, l] * ϵ̂ᵢ[l]
        end

        Hᵢmle = ∂gᵢ' * Σinvs[i] * ∂gᵢ

        Hmle[
            slice(τᵢ, measured_components, zdim),
            slice(τᵢ, measured_components, zdim)
        ] = sparse(Hᵢmle)

        ∇ᵢmle = ΔY.ys[i]' * Σinvs[i] * ∂gᵢ

        if i == M
            Hᵢmle *= Qyf
            ∇ᵢmle *= Qyf
        else
            Hᵢmle *= Qy
            ∇ᵢmle *= Qy
        end

        ∇[slice(τᵢ, measured_components, zdim)] = ∇ᵢmle
    end

    return Hmle, ∇
end



"""
    build_mle_hessian_and_gradient()

Builds the hessian matrix and gradient vector for the MLE objective term.
"""
@inline function build_mle_hessian_and_gradient(
    ΔY::MeasurementData,
    Z::NamedTrajectory,
    ∂g::Function,
    ∂²g::Function,
    Qy::Float64,
    Qyf::Float64,
    Σinvs::Vector{<:AbstractMatrix},
    dims::NamedTuple,
    measured_components::AbstractVector{Int},
    T::Int, # number of time steps
    correction_term::Bool
)
    M = length(ΔY.τs)
    zdim = dims.states + dims.controls

    Hmle = spzeros(zdim * T, zdim * T)

    ∇ = zeros(zdim * T)

    for i = 1:M

        τᵢ = ΔY.τs[i]
        yᵢ = ΔY.ys[i]
        ψ̃ᵢ = Z.ψ̃[:, τᵢ]

        ∂gᵢ = ∂g(ψ̃ᵢ)

        if correction_term
            ∂²gᵢ = ∂²g(ψ̃ᵢ)
            ϵ̂ᵢ = pinv(∂gᵢ) * yᵢ
            @einsum ∂gᵢ[j, k] += ∂²gᵢ[j, k, l] * ϵ̂ᵢ[l]
        end

        Hᵢmle = ∂gᵢ' * Σinvs[i] * ∂gᵢ

        Hmle[
            slice(τᵢ, measured_components, zdim),
            slice(τᵢ, measured_components, zdim)
        ] = sparse(Hᵢmle)

        ∇ᵢmle = yᵢ' * Σinvs[i] * ∂gᵢ

        if i == M
            Hᵢmle *= Qyf
            ∇ᵢmle *= Qyf
        else
            Hᵢmle *= Qy
            ∇ᵢmle *= Qy
        end

        ∇[slice(τᵢ, measured_components, zdim)] = ∇ᵢmle
    end

    return Hmle, ∇
end


"""
    build_initial_and_final_constraint_matrix(Z::NamedTrajectory)

Builds and vcats the matrices to pick out the initial and final states of constrained components.
"""
@inline function build_initial_and_final_constraint_matrix(Z::NamedTrajectory)

    initCs = []

    for key ∈ keys(Z.initial)
        # ignore the wavefunction state constraints
        if key == :ψ̃
            continue
        end
        C = spzeros(Z.dims[key], Z.dim * Z.T)
        C[
            1:Z.dims[key],
            slice(1, Z.components[key], Z.dim)
        ] = sparse(I(Z.dims[key]))
        push!(initCs, C)
    end

    finalCs = []

    for key ∈ keys(Z.final)
        if key == :ψ̃
            continue
        end
        C = spzeros(Z.dims[key], Z.dim * Z.T)
        C[
            1:Z.dims[key],
            slice(Z.T, Z.components[key], Z.dim)
        ] = sparse(I(Z.dims[key]))
        push!(finalCs, C)
    end

    Cs = [initCs; finalCs]

    if isempty(Cs)
        return spzeros(Float64, 0, Z.dim * Z.T)
    else
        return sparse_vcat(Cs...)
    end
end

"""
    build_initial_and_final_constraints(Z::NamedTrajectory)

Build the constraint vector for the initial and final states of constrained components.
"""
@inline function build_initial_and_final_constraints(Z::NamedTrajectory)
    if !isempty(keys(Z.initial))
        # initial_cons = vcat([Z.initial[key] for key ∈ keys(Z.initial) if key != :ψ̃]...)
        initial_cons = zeros(sum([Z.dims[key] for key ∈ keys(Z.initial) if key != :ψ̃]))
    else
        initial_cons = zeros(0)
    end
    if !isempty(keys(Z.final))
        # final_cons = vcat([Z.final[key] for key ∈ keys(Z.final) if key != :ψ̃]...)
        final_cons = zeros(sum([Z.dims[key] for key ∈ keys(Z.final) if key != :ψ̃]))
    else
        final_cons = zeros(0)
    end
    return vcat(initial_cons, final_cons)
end

"""
    build_initial_wavefunction_constraint_matrix(Z::NamedTrajectory)

Builds the matrix to pick out the initial wavefunction state.
"""
@inline function build_initial_wavefunction_constraint_matrix(Z::NamedTrajectory)
    C = spzeros(Z.dims[:ψ̃], Z.dim * Z.T)
    C[
        1:Z.dims[:ψ̃],
        slice(1, Z.components[:ψ̃], Z.dim)
    ] = sparse(I(Z.dims[:ψ̃]))
    return C
end


"""
    build_bounds_constraint_matrix(Z::NamedTrajectory)

Builds the bounds matrix by vcating the matrices that pick out the states/controls that are bounded in `Z`.

Acts on all timeslices: `t ∈ 1:T`
"""
@inline function build_bounds_constraint_matrix(Z::NamedTrajectory)
    Cs = []
    for key ∈ keys(Z.bounds)
        C = spzeros(Z.dims[key] * Z.T, Z.dim * Z.T)
        println("building constraint matrix for $key")
        t_start = time()
        for t = 1:Z.T
            C[
                slice(t, Z.dims[key]),
                slice(t, Z.components[key], Z.dim)
            ] = sparse(I(Z.dims[key]))

            if t % 100 == 0
                t_elapsed = time() - t_start
                t_remaining = t_elapsed / t * (Z.T - t)
                println("t = $t")
                # show estimeated time remaining in hours with 2 decimal places
                println("estimated time remaining: $(round(t_remaining / 3600, digits=2)) hours")

            end
        end
        push!(Cs, C)
    end
    if isempty(Cs)
        return spzeros(Float64, 0, Z.dim * Z.T)
    else
        return sparse_vcat(Cs...)
    end
end

"""
    build_bounds_constraints(Z::NamedTrajectory)

Builds the bounds vector by vcating the bounds vectors for each bounded componenent in `Z`.

Acts on all timeslices: `t ∈ 1:T`
"""
@inline function build_bounds_constraints(Z::NamedTrajectory)
    b_lbs = []
    b_ubs = []
    for key ∈ keys(Z.bounds)

        b_lb = vcat(fill(Z.bounds[key][1], Z.T)...) - vec(Z[key][:, 1:Z.T])

        b_ub = vcat(fill(Z.bounds[key][2], Z.T)...) - vec(Z[key][:, 1:Z.T])

        push!(b_lbs, b_lb)
        push!(b_ubs, b_ub)
    end
    if isempty(b_lbs)
        return zeros(0), zeros(0)
    else
        return vcat(b_lbs...), vcat(b_ubs...)
    end
end


"""
    build_constraints(Z::NamedTrajectory)

Builds the entire constraint vector -- i.e., dynamics, bounds, and initial/final constraints.
"""
@inline function build_constraints(Z::NamedTrajectory)
    ∂F_cons = zeros(Z.dims.states * (Z.T - 1))
    B_lb, B_ub = build_bounds_constraints(Z)
    C_init_final = build_initial_and_final_constraints(Z)
    C_ψ̃₁ = zeros(Z.dims.ψ̃)
    lb = vcat(∂F_cons, B_lb, C_init_final, C_ψ̃₁)
    ub = vcat(∂F_cons, B_ub, C_init_final, C_ψ̃₁)
    return lb, ub
end


"""
    build_dynamics_constraint_jacobian(Ẑ::NamedTrajectory, ∂f::Function)

Builds the Jacobian of the dynamics constraint for the given trajectory
"""
@views function build_dynamics_constraint_jacobian(
    Ẑ::NamedTrajectory,
    ∂f::Function
)
    ∂F = zeros(
        Ẑ.dims.states * (Ẑ.T - 1),
        Ẑ.dim * Ẑ.T
    )

    Threads.@threads for t = 1:Ẑ.T - 1

        # if t % 100 == 0
        #     println("Building dynamics constraint Jacobian: t = $(t) / $(Ẑ.T - 1)")
        #     # estimate completion time
        #     t_elapsed = time() - t_start
        #     t_remaining = (Ẑ.T - 1 - t) * t_elapsed / t
        #     # show estimated time remaining in hours rounded to two decimal places
        #     println("Estimated time remaining: $(round(t_remaining / 3600; digits=2)) hours")
        #     println()
        # end

        zₜzₜ₊₁ = vec(Ẑ.data[:, t:t+1])

        ∂F[slice(t, Ẑ.dims.states), slice(t:t+1, Ẑ.dim)] = ∂f(zₜzₜ₊₁)
    end

    return sparse(∂F)
end

end
